/* PKI Resource Query Protocol Message implementation
 * (c) 2004-2007 by Massimiliano Pala and OpenCA Group
 * All Rights Reserved
 *
 * This software is released under the GPL2 License included
 * in the archive. You can not remove this copyright notice.
 */
                                                                                
#include <libpki/pki.h>

/* Signature ::= SEQUENCE {
 *	signatureAlgorithm	AlgorithmIdentifier,
 *	signature		BIT STRING,
 *	certs			[0] EXPLICIT SEQUENCE OF Certificate OPT }
 */

ASN1_SEQUENCE(PRQP_SIGNATURE) = {
	ASN1_SIMPLE(PRQP_SIGNATURE, signatureAlgorithm, X509_ALGOR ),
	ASN1_SIMPLE(PRQP_SIGNATURE, signature, ASN1_BIT_STRING),
	ASN1_SIMPLE(PRQP_SIGNATURE, signerCert, X509),
	ASN1_EXP_SEQUENCE_OF_OPT(PRQP_SIGNATURE, otherCerts, X509, 1)
} ASN1_SEQUENCE_END(PRQP_SIGNATURE)

IMPLEMENT_ASN1_FUNCTIONS(PRQP_SIGNATURE)

/* BasicCertIndentifier ::= SEQUENCE {
 *	serialNumber		CertificateSerialNumber,
 *	issuerNameHash		OCTET STRING	}
 */

/*
ASN1_SEQUENCE(BASIC_CERT_IDENTIFIER) = {
	ASN1_SIMPLE(BASIC_CERT_IDENTIFIER,subjectNameHash,ASN1_OCTET_STRING),
	ASN1_EXP_OPT(BASIC_CERT_IDENTIFIER,issuerNameHash,ASN1_OCTET_STRING,0),
	ASN1_EXP_OPT(BASIC_CERT_IDENTIFIER,serialNumber,ASN1_INTEGER,1)
} ASN1_SEQUENCE_END(BASIC_CERT_IDENTIFIER)
*/

ASN1_SEQUENCE(BASIC_CERT_IDENTIFIER) = {
	ASN1_SIMPLE(BASIC_CERT_IDENTIFIER,issuerNameHash,ASN1_OCTET_STRING),
	ASN1_SIMPLE(BASIC_CERT_IDENTIFIER,serialNumber,ASN1_INTEGER)
} ASN1_SEQUENCE_END(BASIC_CERT_IDENTIFIER)

IMPLEMENT_ASN1_FUNCTIONS(BASIC_CERT_IDENTIFIER)
IMPLEMENT_ASN1_DUP_FUNCTION( BASIC_CERT_IDENTIFIER )

/* ExtendedCertInfo ::= SEQUENCE {
 *	certificateHash		OCTET STRING,
 *	subjectKeyHash		OCTET STRING,
 *	subjectKeyIdentifier	[0] KeyIdentifier	OPTIONAL,
 *	issuerKeyIdentifier	[1] KeyIdentifier	OPTIONAL }
 */

ASN1_SEQUENCE(EXTENDED_CERT_INFO) = {
	ASN1_SIMPLE(EXTENDED_CERT_INFO, certificateHash, ASN1_OCTET_STRING),
	ASN1_SIMPLE(EXTENDED_CERT_INFO, subjectKeyHash, ASN1_OCTET_STRING),
	ASN1_EXP_OPT(EXTENDED_CERT_INFO, subjectKeyId, ASN1_OCTET_STRING, 0),
	ASN1_EXP_OPT(EXTENDED_CERT_INFO, issuerKeyId, ASN1_OCTET_STRING, 1)
} ASN1_SEQUENCE_END(EXTENDED_CERT_INFO)

IMPLEMENT_ASN1_FUNCTIONS( EXTENDED_CERT_INFO )
IMPLEMENT_ASN1_DUP_FUNCTION( EXTENDED_CERT_INFO)

/* CertIdentifier ::= SEQUENCE {
 *	hashAlgorithm		AlgorithmIdentifier,
 *	basicCertIdentifier	BasicCertIdentifier,
 *	extInfo			[0] ExtendedCertInfo	OPTIONAL,
 *	caCertificate		[1] Certificate		OPTIONAL,
 *	issuedCertificate	[2] Certificate		OPTIONAL }
 */

ASN1_SEQUENCE(CERT_IDENTIFIER) = {
	ASN1_SIMPLE(CERT_IDENTIFIER, hashAlgorithm, X509_ALGOR),
	ASN1_SIMPLE(CERT_IDENTIFIER, basicCertId, BASIC_CERT_IDENTIFIER),
	ASN1_EXP_OPT(CERT_IDENTIFIER, extInfo, EXTENDED_CERT_INFO, 0),
	ASN1_EXP_OPT(CERT_IDENTIFIER, caCert, X509, 1),
	ASN1_EXP_OPT(CERT_IDENTIFIER, issuedCert, X509, 2)
} ASN1_SEQUENCE_END(CERT_IDENTIFIER)

IMPLEMENT_ASN1_FUNCTIONS(CERT_IDENTIFIER)
IMPLEMENT_ASN1_DUP_FUNCTION(CERT_IDENTIFIER)

/* ResourceIdentifier ::= SEQUENCE {
 * 	resourceId	OBJECT IDENTIFIER,
 * 	version		[0] INTEGER	OPTIONAL }
 */

ASN1_SEQUENCE(RESOURCE_IDENTIFIER) = {
	ASN1_SIMPLE( RESOURCE_IDENTIFIER, resourceId, ASN1_OBJECT ),
	ASN1_EXP_OPT( RESOURCE_IDENTIFIER, version, ASN1_INTEGER, 0 ),
	ASN1_EXP_OPT( RESOURCE_IDENTIFIER, oid, ASN1_OBJECT, 1 )
} ASN1_SEQUENCE_END ( RESOURCE_IDENTIFIER )

IMPLEMENT_ASN1_FUNCTIONS( RESOURCE_IDENTIFIER )
IMPLEMENT_ASN1_DUP_FUNCTION( RESOURCE_IDENTIFIER )

/* ResourceRequestToken ::= SEQUENCE {
 *	ca		 caCertIdentifier,
 *	resourceList [0] SEQUENCE OF OBJECT IDENTIFIER	OPTIONAL }
 */

ASN1_SEQUENCE(RESOURCE_REQUEST_TOKEN) = {
	ASN1_SIMPLE(RESOURCE_REQUEST_TOKEN, ca, CERT_IDENTIFIER),
	ASN1_EXP_SEQUENCE_OF(RESOURCE_REQUEST_TOKEN, resourceList, RESOURCE_IDENTIFIER, 0)
} ASN1_SEQUENCE_END(RESOURCE_REQUEST_TOKEN)

IMPLEMENT_ASN1_FUNCTIONS(RESOURCE_REQUEST_TOKEN)

/* TBSReqData ::= SEQUENCE {
 * 	version		INTEGER 		{v(1)},
 *	nonce		[0] INTEGER		OPTIONAL,
 *	serviceToken	ResourceRequestToken,
 *	extensions	[1]	IMPLICIT Extensions	OPTIONAL }
*/

ASN1_SEQUENCE(PRQP_TBS_REQ_DATA) = {
	ASN1_SIMPLE(PRQP_TBS_REQ_DATA, version, ASN1_INTEGER),
	ASN1_EXP_OPT(PRQP_TBS_REQ_DATA, nonce, ASN1_INTEGER, 0),
	ASN1_SIMPLE(PRQP_TBS_REQ_DATA, producedAt, ASN1_GENERALIZEDTIME),
	ASN1_SIMPLE(PRQP_TBS_REQ_DATA, serviceToken, RESOURCE_REQUEST_TOKEN),
	ASN1_IMP_SEQUENCE_OF_OPT(PRQP_TBS_REQ_DATA, extensions, X509_EXTENSION, 1)
} ASN1_SEQUENCE_END(PRQP_TBS_REQ_DATA)

IMPLEMENT_ASN1_FUNCTIONS(PRQP_TBS_REQ_DATA)


/* PRQPReqest ::= SEQUENCE {
 *	requestData	TBSReqData,
 *	signature	[0] EXPLICIT Signature OPTIONAL }
 */

ASN1_SEQUENCE(PKI_PRQP_REQ) = {
	ASN1_SIMPLE(PKI_PRQP_REQ, requestData, PRQP_TBS_REQ_DATA ),
	ASN1_EXP_OPT(PKI_PRQP_REQ, prqpSignature, PRQP_SIGNATURE, 0)
} ASN1_SEQUENCE_END(PKI_PRQP_REQ)

IMPLEMENT_ASN1_FUNCTIONS(PKI_PRQP_REQ)

IMPLEMENT_ASN1_DUP_FUNCTION(PKI_PRQP_REQ)

/*
PKI_PRQP_REQ * PKI_PRQP_REQ_dup ( PKI_PRQP_REQ *req ) {
	ASN1_item_dup ( &PKI_PRQP_REQ_it, req );
}
*/

/* end */
